/***************************************************************
 * The code for ZIP-AES with Gray code counter
 ***************************************************************/
#include"define.h"
#include<vector>

using namespace std;

class ZIPAES_GRAY {
public:
  // subkeys for AES128
  __m128i w128[11];

  // constants for gray-code counter
  __m128i A[32];
  __m128i B[32];

  /***************************************************************
   * for debug
   ***************************************************************/
  void dispXMM(__m128i xmm) {

    MIE_ALIGN(128) unsigned char val[16] = { 0 };
    _mm_store_si128((__m128i*)(val), xmm);

    // little endian
    for (int i = 15; i >= 0; i--) {
      printf("%02X", val[i]);
    }
    printf("\n");
  }

  /***************************************************************
   * setup
   *  aesKeySchedule : Key schedule algorithm for AES
   *  programSetUp : set up (key schedule & gray-code counter)
   ***************************************************************/
  void aesKeySchedule(const unsigned char* k) {
    __m128i work = _mm_loadu_si128(reinterpret_cast<const __m128i*>(k));  // work = w3 : w2 : w1 : w0
    w128[0] = work;

    __m128i t, t2;

    #define EXPAND(n, RCON) \
        t  = _mm_slli_si128(work, 4);       /* t    = w2                : w1             : w0          : 0        */\
        t  = _mm_xor_si128(work, t);        /* t    = w3^w2             : w2^w1          : w1^w0       : w0       */\
        t2 = _mm_slli_si128(t, 8);          /* t2   = w1^w0             : w0             : 0           : 0        */\
        t  = _mm_xor_si128(t, t2);          /* t    = w3^w2^w1^w0       : w2^w1^w0       : w1^w0       : w0       */\
        work = _mm_aeskeygenassist_si128(work, RCON);           /* work = f(w3) : - : - : - */                      \
        work = _mm_shuffle_epi32(work, 0xFF);/*work = f(w3)             : f(w3)          : f(w3)       ; f(w3)    */\
        work = _mm_xor_si128(t, work);      /* work = w3^w2^w1^w0^f(w3) : w2^w1^w0^f(w3) : w1^w0^f(w3) : w0^f(w3) */\
        w128[n] = work;                     /* work = w7 : w6 : w5 : w4 */

    //
    EXPAND(1, 0x01);
    EXPAND(2, 0x02);
    EXPAND(3, 0x04);
    EXPAND(4, 0x08);
    EXPAND(5, 0x10);
    EXPAND(6, 0x20);
    EXPAND(7, 0x40);
    EXPAND(8, 0x80);
    EXPAND(9, 0x1b);
    EXPAND(10, 0x36);


    //w128[6] = _mm_aesimc_si128(w128[6]);
    w128[7] = _mm_aesimc_si128(w128[7]);
    w128[8] = _mm_aesimc_si128(w128[8]);
    w128[9] = _mm_aesimc_si128(w128[9]);
    w128[10] = _mm_aesimc_si128(w128[10]);
  }
  void programSetUp(const unsigned char* k) {
    // run aes key scheulde and store subkey into w128
    aesKeySchedule(k);  
    
    // generate constants for gray-code counter
    for (int i = 0; i < 32; i++) {

      __m128i x = _mm_cvtsi64_si128(1 << i);
      __m128i y = _mm_aesimc_si128(x);

      A[i] = x;
      B[i] = y;
    }
  }


  /***********************************************************************************
   * core
   *  aes : single block
   *  aes_double : 2 block parallel
   *  aes_quad   : 4 block parallel
   ***********************************************************************************/
  inline void aes(__m128i stL, __m128i stR, __m128i& ks) {

    stL = _mm_aesenc_si128(stL, w128[1]);
    stR = _mm_aesdec_si128(stR, w128[7]);

    stL = _mm_aesenc_si128(stL, w128[2]);
    stR = _mm_aesdec_si128(stR, w128[8]);

    stL = _mm_aesenc_si128(stL, w128[3]);
    stR = _mm_aesdec_si128(stR, w128[9]);

    stL = _mm_aesenc_si128(stL, w128[4]);
    stR = _mm_aesdec_si128(stR, w128[10]);

    stL = _mm_aesenc_si128(stL, w128[5]);
    ks = _mm_aesdeclast_si128(stR, stL);

  }
  inline void aes_double(__m128i st0L, __m128i st0R, __m128i st1L, __m128i st1R, __m128i& ks0, __m128i& ks1) {

    st0L = _mm_aesenc_si128(st0L, w128[1]);
    st1L = _mm_aesenc_si128(st1L, w128[1]);

    st0R = _mm_aesdec_si128(st0R, w128[7]);
    st1R = _mm_aesdec_si128(st1R, w128[7]);

    st0L = _mm_aesenc_si128(st0L, w128[2]);
    st1L = _mm_aesenc_si128(st1L, w128[2]);

    st0R = _mm_aesdec_si128(st0R, w128[8]);
    st1R = _mm_aesdec_si128(st1R, w128[8]);

    st0L = _mm_aesenc_si128(st0L, w128[3]);
    st1L = _mm_aesenc_si128(st1L, w128[3]);

    st0R = _mm_aesdec_si128(st0R, w128[9]);
    st1R = _mm_aesdec_si128(st1R, w128[9]);

    st0L = _mm_aesenc_si128(st0L, w128[4]);
    st1L = _mm_aesenc_si128(st1L, w128[4]);

    st0R = _mm_aesdec_si128(st0R, w128[10]);
    st1R = _mm_aesdec_si128(st1R, w128[10]);

    st0L = _mm_aesenc_si128(st0L, w128[5]);
    st1L = _mm_aesenc_si128(st1L, w128[5]);

    ks0 = _mm_aesdeclast_si128(st0R, st0L);
    ks1 = _mm_aesdeclast_si128(st1R, st1L);

  }
  inline void aes_quad(__m128i st0L, __m128i st0R, __m128i st1L, __m128i st1R, __m128i st2L, __m128i st2R, __m128i st3L, __m128i st3R, __m128i& ks0, __m128i& ks1, __m128i& ks2, __m128i& ks3) {


    st0L = _mm_aesenc_si128(st0L, w128[1]);
    st1L = _mm_aesenc_si128(st1L, w128[1]);
    st2L = _mm_aesenc_si128(st2L, w128[1]);
    st3L = _mm_aesenc_si128(st3L, w128[1]);

    st0R = _mm_aesdec_si128(st0R, w128[7]);
    st1R = _mm_aesdec_si128(st1R, w128[7]);
    st2R = _mm_aesdec_si128(st2R, w128[7]);
    st3R = _mm_aesdec_si128(st3R, w128[7]);

    st0L = _mm_aesenc_si128(st0L, w128[2]);
    st1L = _mm_aesenc_si128(st1L, w128[2]);
    st2L = _mm_aesenc_si128(st2L, w128[2]);
    st3L = _mm_aesenc_si128(st3L, w128[2]);

    st0R = _mm_aesdec_si128(st0R, w128[8]);
    st1R = _mm_aesdec_si128(st1R, w128[8]);
    st2R = _mm_aesdec_si128(st2R, w128[8]);
    st3R = _mm_aesdec_si128(st3R, w128[8]);

    st0L = _mm_aesenc_si128(st0L, w128[3]);
    st1L = _mm_aesenc_si128(st1L, w128[3]);
    st2L = _mm_aesenc_si128(st2L, w128[3]);
    st3L = _mm_aesenc_si128(st3L, w128[3]);

    st0R = _mm_aesdec_si128(st0R, w128[9]);
    st1R = _mm_aesdec_si128(st1R, w128[9]);
    st2R = _mm_aesdec_si128(st2R, w128[9]);
    st3R = _mm_aesdec_si128(st3R, w128[9]);

    st0L = _mm_aesenc_si128(st0L, w128[4]);
    st1L = _mm_aesenc_si128(st1L, w128[4]);
    st2L = _mm_aesenc_si128(st2L, w128[4]);
    st3L = _mm_aesenc_si128(st3L, w128[4]);

    st0R = _mm_aesdec_si128(st0R, w128[10]);
    st1R = _mm_aesdec_si128(st1R, w128[10]);
    st2R = _mm_aesdec_si128(st2R, w128[10]);
    st3R = _mm_aesdec_si128(st3R, w128[10]);

    st0L = _mm_aesenc_si128(st0L, w128[5]);
    st1L = _mm_aesenc_si128(st1L, w128[5]);
    st2L = _mm_aesenc_si128(st2L, w128[5]);
    st3L = _mm_aesenc_si128(st3L, w128[5]);

    ks0 = _mm_aesdeclast_si128(st0R, st0L);
    ks1 = _mm_aesdeclast_si128(st1R, st1L);
    ks2 = _mm_aesdeclast_si128(st2R, st2L);
    ks3 = _mm_aesdeclast_si128(st3R, st3L);

  }

};

ZIPAES_GRAY zipaes_gray;

void zipaes_gray_set_up(const unsigned char* k) {
  zipaes_gray.programSetUp(k);
}

//
void zipaes_gray_enc(
  const unsigned char* _K,     // Secret key                
  const unsigned char* _N,     // Nonce                     
  const unsigned char* _M,     // Message                   
  unsigned long long   _mlen,  // Length of Message         
  unsigned char* C_            // Ciphertext              
) {

  // set
  __m128i ctr = _mm_load_si128((const __m128i*)(_N));

  // set
  __m128i ctrL0 = _mm_xor_si128(ctr, zipaes_gray.w128[0]);
  __m128i ctrR0 = _mm_xor_si128(ctr, zipaes_gray.w128[6]);
  ctrR0 = _mm_aesimc_si128(ctrR0);

  int cnt = 0;
  long long m_p = 0;
  __m128i ks0, ks1, ks2, ks3, ks4, ks5, ks6, ks7;


  while (m_p + 112 < (long long)(_mlen)) {

    // 1st loop
    __m128i ctrL1 = _mm_xor_si128(ctrL0, zipaes_gray.A[0]);
    __m128i ctrR1 = _mm_xor_si128(ctrR0, zipaes_gray.B[0]);

    __m128i ctrL2 = _mm_xor_si128(ctrL1, zipaes_gray.A[1]);
    __m128i ctrR2 = _mm_xor_si128(ctrR1, zipaes_gray.B[1]);

    __m128i ctrL3 = _mm_xor_si128(ctrL2, zipaes_gray.A[0]);
    __m128i ctrR3 = _mm_xor_si128(ctrR2, zipaes_gray.B[0]);

    // ks = enc(iv)
    zipaes_gray.aes_quad(ctrL0, ctrR0, ctrL1, ctrR1, ctrL2, ctrR2, ctrL3, ctrR3, ks0, ks1, ks2, ks3);

    __m128i ctrL4 = _mm_xor_si128(ctrL3, zipaes_gray.A[2]);
    __m128i ctrR4 = _mm_xor_si128(ctrR3, zipaes_gray.B[2]);

    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    __m128i m2 = _mm_load_si128((const __m128i*)(_M + m_p));
    m2 = _mm_xor_si128(ks2, m2);
    _mm_store_si128((__m128i*)(C_ + m_p), m2);
    m_p += 16;

    __m128i m3 = _mm_load_si128((const __m128i*)(_M + m_p));
    m3 = _mm_xor_si128(ks3, m3);
    _mm_store_si128((__m128i*)(C_ + m_p), m3);
    m_p += 16;







    // 2nd loop
    __m128i ctrL5 = _mm_xor_si128(ctrL4, zipaes_gray.A[0]);
    __m128i ctrR5 = _mm_xor_si128(ctrR4, zipaes_gray.B[0]);

    __m128i ctrL6 = _mm_xor_si128(ctrL5, zipaes_gray.A[1]);
    __m128i ctrR6 = _mm_xor_si128(ctrR5, zipaes_gray.B[1]);

    __m128i ctrL7 = _mm_xor_si128(ctrL6, zipaes_gray.A[0]);
    __m128i ctrR7 = _mm_xor_si128(ctrR6, zipaes_gray.B[0]);

    // ks = enc(iv)
    zipaes_gray.aes_quad(ctrL4, ctrR4, ctrL5, ctrR5, ctrL6, ctrR6, ctrL7, ctrR7, ks4, ks5, ks6, ks7);

    // update gray
    cnt += 8;
    int index = __builtin_ctz(cnt);

    //
    __m128i m4 = _mm_load_si128((const __m128i*)(_M + m_p));
    m4 = _mm_xor_si128(ks4, m4);
    _mm_store_si128((__m128i*)(C_ + m_p), m4);
    m_p += 16;

    __m128i m5 = _mm_load_si128((const __m128i*)(_M + m_p));
    m5 = _mm_xor_si128(ks5, m5);
    _mm_store_si128((__m128i*)(C_ + m_p), m5);
    m_p += 16;

    __m128i m6 = _mm_load_si128((const __m128i*)(_M + m_p));
    m6 = _mm_xor_si128(ks6, m6);
    _mm_store_si128((__m128i*)(C_ + m_p), m6);
    m_p += 16;

    __m128i m7 = _mm_load_si128((const __m128i*)(_M + m_p));
    m7 = _mm_xor_si128(ks7, m7);
    _mm_store_si128((__m128i*)(C_ + m_p), m7);
    m_p += 16;






    // update
    ctrL0 = _mm_xor_si128(ctrL7, zipaes_gray.A[index]);
    ctrR0 = _mm_xor_si128(ctrR7, zipaes_gray.B[index]);


  }

  while (m_p + 48 < (long long)(_mlen)) {

    __m128i ctrL1 = _mm_xor_si128(ctrL0, zipaes_gray.A[0]);
    __m128i ctrR1 = _mm_xor_si128(ctrR0, zipaes_gray.B[0]);

    __m128i ctrL2 = _mm_xor_si128(ctrL1, zipaes_gray.A[1]);
    __m128i ctrR2 = _mm_xor_si128(ctrR1, zipaes_gray.B[1]);

    __m128i ctrL3 = _mm_xor_si128(ctrL2, zipaes_gray.A[0]);
    __m128i ctrR3 = _mm_xor_si128(ctrR2, zipaes_gray.B[0]);

    // ks = enc(iv)
    zipaes_gray.aes_quad(ctrL0, ctrR0, ctrL1, ctrR1, ctrL2, ctrR2, ctrL3, ctrR3, ks0, ks1, ks2, ks3);

    //
    cnt += 4;
    //int index = __builtin_ctz(cnt);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    // c = m ^ ks
    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    // c = m ^ ks
    __m128i m2 = _mm_load_si128((const __m128i*)(_M + m_p));
    m2 = _mm_xor_si128(ks2, m2);
    _mm_store_si128((__m128i*)(C_ + m_p), m2);
    m_p += 16;

    // c = m ^ ks
    __m128i m3 = _mm_load_si128((const __m128i*)(_M + m_p));
    m3 = _mm_xor_si128(ks3, m3);
    _mm_store_si128((__m128i*)(C_ + m_p), m3);
    m_p += 16;

    //cout << index << " ";
    //
    ctrL0 = _mm_xor_si128(ctrL3, zipaes_gray.A[2]);
    ctrR0 = _mm_xor_si128(ctrR3, zipaes_gray.B[2]);


  }

  while (m_p + 16 < (long long)(_mlen)) {

    __m128i ctrL1 = _mm_xor_si128(ctrL0, zipaes_gray.A[0]);
    __m128i ctrR1 = _mm_xor_si128(ctrR0, zipaes_gray.A[0]);

    // ks = enc(iv)
    zipaes_gray.aes_double(ctrL0, ctrR0, ctrL1, ctrR1, ks0, ks1);

    //
    cnt += 2;
    //int index = __builtin_ctz(cnt);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

    // c = m ^ ks
    __m128i m1 = _mm_load_si128((const __m128i*)(_M + m_p));
    m1 = _mm_xor_si128(ks1, m1);
    _mm_store_si128((__m128i*)(C_ + m_p), m1);
    m_p += 16;

    //cout << index << " ";
    //
    ctrL0 = _mm_xor_si128(ctrL1, zipaes_gray.A[1]);
    ctrR0 = _mm_xor_si128(ctrR1, zipaes_gray.B[1]);


  }

  while (m_p < (long long)(_mlen)) {

    // ks = enc(iv)
    zipaes_gray.aes(ctrL0, ctrR0, ks0);

    // c = m ^ ks
    __m128i m0 = _mm_load_si128((const __m128i*)(_M + m_p));
    m0 = _mm_xor_si128(ks0, m0);
    _mm_store_si128((__m128i*)(C_ + m_p), m0);
    m_p += 16;

  }




}
